/*
Copyright (C) 1996-1997 Id Software, Inc.

This program is free software; you can redistribute it and/or
modify it under the terms of the GNU General Public License
as published by the Free Software Foundation; either version 2
of the License, or (at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  

See the GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program; if not, write to the Free Software
Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.

*/

// lightmap building and rebuilding - kept in a seperate file so that gl_rsurf is more legible...

#include "quakedef.h"
#include "gl_mtex.h"
#include "gl_lightmap.h"

#ifndef GL_RGBA4
#define	GL_RGBA4	0x8056
#endif

// 0 = card has negative lighting, 1 = card lights normally
int neglight;

int		lightmap_bytes;		// 1, 2, or 4

int		lightmap_textures;

unsigned		blocklights[BLOCK_WIDTH * BLOCK_HEIGHT];
unsigned		blocklightcolours[3][BLOCK_WIDTH * BLOCK_HEIGHT];

qboolean	lightmap_modified[MAX_LIGHTMAPS];
glRect_t	lightmap_rectchange[MAX_LIGHTMAPS];

int			allocated[MAX_LIGHTMAPS][BLOCK_WIDTH];

byte lightmaps[MAX_LIGHTMAPS*BLOCK_WIDTH*BLOCK_HEIGHT*4];

void GL_BuildVertexLightColour (glmsurface_t *fa);
void GL_BuildVertexLightWhite (glmsurface_t *fa);


extern cvar_t	gl_vertexlight;
extern int gl_useclight;


int bound (int lower, int num, int upper)
{
	if (num < lower) return lower;
	if (num > upper) return upper;
	return num;
}


/*
===============
R_AddDynamicLights
===============
*/
void R_AddDynamicLights (glmsurface_t *surf)
{
	int			lnum;
	int			sd, td;
	float		dist, rad, minlight;
	float		rad_minus_dist;
	vec3_t		impact, local;
	int			s, t;
	glmtexinfo_t	*tex;
	dlight_t *dl;

	// mh - major speed increase coming up here - yowsa!
	int t_x_smax;
	float dl_red;
	float dl_green;
	float dl_blue;

	if (surf->BadExtents) return;

	tex = surf->texinfo;

	for (lnum=0 ; lnum<MAX_DLIGHTS ; lnum++)
	{
		dl = &cl_dlights[lnum];

		// check for dead lights - this shouldn't really be necessary, 
		// but i've learned (the hard way) not to trust id's code
		if (dl->die < cl.time || !dl->radius)
			continue;

		if ( !(surf->dlightbits & (1<<lnum) ) )
			continue;		// not lit by this light

		rad = cl_dlights[lnum].radius;

		dist = DotProduct (cl_dlights[lnum].origin, surf->plane->normal) - surf->plane->dist;

		rad -= fabs(dist);
		minlight = cl_dlights[lnum].minlight;

		if (rad < minlight)
			continue;

		dl_red = dl->colour[0] * 256;
		dl_green = dl->colour[1] * 256;
		dl_blue = dl->colour[2] * 256;

		minlight = rad - minlight;

		impact[0] = cl_dlights[lnum].origin[0] - surf->plane->normal[0]*dist;
		impact[1] = cl_dlights[lnum].origin[1] - surf->plane->normal[1]*dist;
		impact[2] = cl_dlights[lnum].origin[2] - surf->plane->normal[2]*dist;

		local[0] = DotProduct (impact, tex->vecs[0]) + tex->vecs[0][3];
		local[1] = DotProduct (impact, tex->vecs[1]) + tex->vecs[1][3];

		local[0] -= surf->texturemins[0];
		local[1] -= surf->texturemins[1];
		
		for (t = 0 ; t<surf->tmax ; t++)
		{
			td = local[1] - t*16;

			if (td < 0)
				td = -td;

			t_x_smax = t * surf->smax;

			for (s=0 ; s<surf->smax ; s++)
			{
				sd = local[0] - s*16;

				if (sd < 0)
					sd = -sd;

				if (sd > td)
					dist = sd + (td>>1);
				else
					dist = td + (sd>>1);

				rad_minus_dist = rad - dist;

				if (dist < minlight && gl_useclight)
				{
					if (!dl->colour[0] && !dl->colour[1] && !dl->colour[2]) 
					{
						// blocklightcolours[2] is not increased here...
						blocklightcolours[0][t_x_smax + s] += rad_minus_dist * 51.2;
						blocklightcolours[1][t_x_smax + s] += rad_minus_dist * 25.6;
					}
					else 
					{
						if (dl_red) blocklightcolours[0][t_x_smax + s] += rad_minus_dist * dl_red;
						if (dl_green) blocklightcolours[1][t_x_smax + s] += rad_minus_dist * dl_green;
						if (dl_blue) blocklightcolours[2][t_x_smax + s] += rad_minus_dist * dl_blue;
					}
				}
				else if (dist < minlight) blocklights[t_x_smax + s] += rad_minus_dist * 256;
			}
		}
	}
}


void R_AddDynamicVertexLights (glmsurface_t *surf)
{
	int lnum, i;
	float dist, rad, minlight;
	dlight_t *dl;
	glpoly_t *p;
	float *v;
	vec3_t dl_dist;
	float		rad_minus_dist;
	float		white_light;

	for (lnum=0 ; lnum<MAX_DLIGHTS ; lnum++)
	{
		dl = &cl_dlights[lnum];

		// check for dead lights - this shouldn't really be necessary, 
		// but i've learned (the hard way) not to trust id's code
		if (dl->die < cl.time || !dl->radius)
			continue;

		if (!(surf->dlightbits & (1<<lnum)))
			continue;		// not lit by this light

		dist = DotProduct (cl_dlights[lnum].origin, surf->plane->normal) - surf->plane->dist;
		rad = cl_dlights[lnum].radius;

		rad -= fabs (dist);
		minlight = cl_dlights[lnum].minlight;

		if (rad < minlight)
			continue;

		// white light...
		if (!gl_useclight)
		{
			white_light = (dl->colour[0] + dl->colour[1] + dl->colour[2]) / 3;

			dl->colour[0] = white_light;
			dl->colour[1] = white_light;
			dl->colour[2] = white_light;
		}

		minlight = rad - minlight;

		for (p = surf->polys; p; p = p->next)
		{
			for (i = 0, v = p->verts[0]; i < p->numverts; i++, v += VERTEXSIZE)
			{
				VectorSubtract (v, dl->origin, dl_dist);
				dist = VLength (dl_dist);

				// 64 is a good all-round value that gives results consistent in appearance with
				// the lightmap method.  these dynamic lights have linear falloff.  perhaps
				// exponential might look better but it would be a little slower...
				rad_minus_dist = (rad - dist) / 64;

				if (dist < minlight)
				{
					// modify the vertex light values - no need to check for coloured light here
					// cos it's all the same anyway - this will give coloured dynamic lights in
					// maps which are loaded with luminance lightmaps!!!  yowsa!!!
					if (!dl->colour[0] && !dl->colour[1] && !dl->colour[2] && gl_useclight)
					{
						v[8] = v[8] + rad_minus_dist * 0.2;
						v[9] = v[9] + rad_minus_dist * 0.1;
					}
					else if (!dl->colour[0] && !dl->colour[1] && !dl->colour[2])
					{
						v[8] = v[8] + rad_minus_dist * 0.1;
						v[9] = v[9] + rad_minus_dist * 0.1;
						v[10] = v[10] + rad_minus_dist * 0.1;
					}
					else
					{
						v[8] = v[8] + rad_minus_dist * dl->colour[0];
						v[9] = v[9] + rad_minus_dist * dl->colour[1];
						v[10] = v[10] + rad_minus_dist * dl->colour[2];
					}
				}
			}
		}
	}
}


/*
===============
R_BuildLightMap

Combine and scale multiple lightmaps into the 8.8 format in blocklights

Both Tomaz and LordHavoc are doing this using pointer arithmetic, so that
must be the right way to go...

UPDATE!
The tomaz/lh code causes coloured lightmaps to overbright somewhat viciously in my engine,
so I've restored my original code.  Pity...

Good job actually, since my code seems to give an extra 2 fps with coloured light
===============
*/
void R_BuildLightMap (glmsurface_t *surf, byte *dest, int stride)
{
    int t, r, s, q;
	int			i, j, size;
	byte		*lightmap;
	unsigned	scale;
	int			maps;
	unsigned	*bl;
    unsigned *blcr, *blcg, *blcb;

	surf->cached_dlight = (surf->dlightframe == r_framecount);

	size = surf->smax * surf->tmax;
	lightmap = surf->samples;

	// set to full bright if no light data
	if (!cl.worldmodel->lightdata)
		for (i=0 ; i<size ; i++)
			if (gl_useclight)
	            blocklightcolours[0][i] = blocklightcolours[1][i] = blocklightcolours[2][i] = 65280;
			else blocklights[i] = 65280;
	else
	{
		// clear to no light
		for (i=0 ; i<size ; i++)
			if (gl_useclight)
		        blocklightcolours[0][i] = blocklightcolours[1][i] = blocklightcolours[2][i] = 0;
			else blocklights[i] = 0;

		// add all the lightmaps
		if (lightmap && gl_useclight)
		{
	        for (maps = 0 ; maps < MAXLIGHTMAPS && surf->styles[maps] != 255 ; maps++)
	        {
	            scale = d_lightstylevalue[surf->styles[maps]];
	            surf->cached_light[maps] = scale; // 8.8 fraction

				for (i=0, j=0 ; i<size ; i++)
				{
					// we're still loading 4 component lightmaps - oh dear...
					blocklightcolours[0][i] += lightmap[j] * scale; j++;
					blocklightcolours[1][i] += lightmap[j] * scale; j++;
					blocklightcolours[2][i] += lightmap[j] * scale; j+=2;
				}

				lightmap += size * 4;
			}
		}
		else if (lightmap)
		{
			for (maps = 0 ; maps < MAXLIGHTMAPS && surf->styles[maps] != 255 ; maps++)
			{
				scale = d_lightstylevalue[surf->styles[maps]];
				surf->cached_light[maps] = scale;	// 8.8 fraction

				for (i=0 ; i<size ; i++)
					blocklights[i] += lightmap[i] * scale;

				lightmap += size;
			}
		}

		// add all the dynamic lights
		if ((surf->dlightframe == r_framecount) && r_dynamic.value) R_AddDynamicLights (surf);
	}

	// bound, invert, and shift
	if (gl_useclight)
	{
		stride -= (surf->smax<<2);

		blcr = blocklightcolours[0];
		blcg = blocklightcolours[1];
		blcb = blocklightcolours[2];

		for (i=0 ; i<surf->tmax ; i++, dest += stride)
		{
			for (j=0 ; j<surf->smax ; j++)
			{
				q = *blcr++; q >>= 7;
				r = *blcg++; r >>= 7;
				s = *blcb++; s >>= 7;

				if (q > 255) q = 255;
				if (r > 255) r = 255;
				if (s > 255) s = 255; 

				// mh - fix for negative lighting on some 3d cards
				if (neglight) dest[0] = 255 - q; else dest[0] = q;
				if (neglight) dest[1] = 255 - r; else dest[1] = r;
				if (neglight) dest[2] = 255 - s; else dest[2] = s;

				dest += lightmap_bytes;
			}
		}
	}
	else
	{
		bl = blocklights;

		for (i=0 ; i<surf->tmax ; i++, dest += stride)
		{
			for (j=0 ; j<surf->smax ; j++)
			{
				t = *bl++; t >>= 7;
				if (t > 255) t = 255;

				if (neglight) dest[j] = 255 - t; else dest[j] = t;
			}
		}
	}
}


// build dynamic lights for vertex lighting
void R_RenderDynamicVertexLightmaps (glmsurface_t *fa)
{
	// check for lightmap modification
	// do this the longer (but faster) way...
	if (fa->styles[0] != 255)
	{
		if (d_lightstylevalue[fa->styles[0]] != fa->cached_light[0]) goto vdynamic;

		if (fa->styles[1] != 255)
		{
			if (d_lightstylevalue[fa->styles[1]] != fa->cached_light[1]) goto vdynamic;

			if (fa->styles[2] != 255)
			{
				if (d_lightstylevalue[fa->styles[2]] != fa->cached_light[2]) goto vdynamic;

				if (fa->styles[3] != 255)
					if (d_lightstylevalue[fa->styles[3]] != fa->cached_light[3]) goto vdynamic;
			}
		}
	}

	if ((fa->dlightframe == r_framecount || fa->cached_dlight) && r_dynamic.value)
	{
vdynamic:
		// store out the new cached lightstyle
		fa->cached_light[0] = d_lightstylevalue[fa->styles[0]];
		fa->cached_light[1] = d_lightstylevalue[fa->styles[1]];
		fa->cached_light[2] = d_lightstylevalue[fa->styles[2]];
		fa->cached_light[3] = d_lightstylevalue[fa->styles[3]];
		fa->cached_dlight = (fa->dlightframe == r_framecount);

		// rebuild the vertex light
		if (gl_useclight)
			GL_BuildVertexLightColour (fa);
		else GL_BuildVertexLightWhite (fa);

		// add dynamic lights to the vertex light
		if ((fa->dlightframe == r_framecount) && r_dynamic.value) R_AddDynamicVertexLights (fa);
	}
}


void R_RenderDynamicWaterLightmaps (glmsurface_t *fa)
{
	// NONE OF THIS WORKS!  dynamic water lighting is a no-go for now, so just don't
	// even bother - we'll try to fix it for the next release.
//	fa->cached_dlight = (fa->dlightframe == r_framecount);

	// no lightstyles hitting these surfs, so leave them alone for now - we might make
	// lightstyles an option later on...
//	if ((fa->dlightframe == r_framecount || fa->cached_dlight) && r_dynamic.value)
//		R_AddDynamicVertexLights (fa);
}


/*
================
R_RenderDynamicLightmaps

================
*/
void R_RenderDynamicLightmaps (glmsurface_t *fa)
{
	int			i;
	glRect_t    *theRect;

	// mh - since alpha test polys use vertex lighting, they must also be done this way.
	if (gl_vertexlight.value || (fa->flags & SURF_TRANS))
	{
		R_RenderDynamicVertexLightmaps (fa);
		return;
	}

	// check for lightmap modification
	// do this the longer (but faster) way...
	if (fa->styles[0] != 255)
	{
		if (d_lightstylevalue[fa->styles[0]] != fa->cached_light[0]) goto dynamic;

		if (fa->styles[1] != 255)
		{
			if (d_lightstylevalue[fa->styles[1]] != fa->cached_light[1]) goto dynamic;

			if (fa->styles[2] != 255)
			{
				if (d_lightstylevalue[fa->styles[2]] != fa->cached_light[2]) goto dynamic;

				if (fa->styles[3] != 255)
					if (d_lightstylevalue[fa->styles[3]] != fa->cached_light[3]) goto dynamic;
			}
		}
	}

	if ((fa->dlightframe == r_framecount || fa->cached_dlight) && r_dynamic.value)
	{
dynamic:
		// somewhat optimised by precalculating a lot of the stuff in here...
		i = fa->lightmaptexturenum;
		lightmap_modified[i] = true;
		theRect = &lightmap_rectchange[i];

		if (fa->light_t < theRect->t) 
		{
			if (theRect->h)
				theRect->h += theRect->t - fa->light_t;

			theRect->t = fa->light_t;
		}

		if (fa->light_s < theRect->l) 
		{
			if (theRect->w)
				theRect->w += theRect->l - fa->light_s;

			theRect->l = fa->light_s;
		}

		if ((theRect->w + theRect->l) < (fa->light_s + fa->smax))
			theRect->w = (fa->light_s-theRect->l)+ fa->smax;

		if ((theRect->h + theRect->t) < (fa->light_t + fa->tmax))
			theRect->h = (fa->light_t-theRect->t)+fa->tmax;

		R_BuildLightMap (fa, fa->base, BLOCK_WIDTH*lightmap_bytes);
	}
}


/*
=============================================================================

  LIGHTMAP ALLOCATION

=============================================================================
*/

// returns a texture number and the position inside it
int AllocBlock (int w, int h, int *x, int *y)
{
	int		i, j;
	int		best, best2;
	int		texnum;

	for (texnum=0 ; texnum<MAX_LIGHTMAPS ; texnum++)
	{
		best = BLOCK_HEIGHT;

		for (i=0 ; i<BLOCK_WIDTH-w ; i++)
		{
			best2 = 0;

			for (j=0 ; j<w ; j++)
			{
				if (allocated[texnum][i+j] >= best)
					break;
				if (allocated[texnum][i+j] > best2)
					best2 = allocated[texnum][i+j];
			}
			if (j == w)
			{	// this is a valid spot
				*x = i;
				*y = best = best2;
			}
		}

		if (best + h > BLOCK_HEIGHT)
			continue;

		for (i=0 ; i<w ; i++)
			allocated[texnum][*x + i] = best + h;

		return texnum;
	}

	Sys_Error ("AllocBlock: full");

	return 0;
}


mvertex_t	*r_pcurrentvertbase;
model_t		*currentmodel;

int	nColinElim;

/*
================
BuildSurfaceDisplayList
This doesn't really belong with lightmap generation code, but it's the handiest place to put it.
================
*/
void BuildSurfaceDisplayList (glmsurface_t *fa)
{
	int			i, lindex, lnumverts;
	medge_t		*pedges, *r_pedge;
	int			vertpage;
	float		*vec;
	float		s, t;
	glpoly_t	*poly;

	// should never ever happen - but can we be too careful here?
	// only check water, cos skies will have standard polys as well as warp_polys.
	if (fa->texinfo->texture->name[0] == '*') return;

	pedges = currentmodel->edges;
	lnumverts = fa->numedges;
	vertpage = 0;

	// draw texture
	poly = Hunk_Alloc (sizeof(glpoly_t) + (lnumverts-4) * VERTEXSIZE*sizeof(float));
	poly->next = fa->polys;
	poly->flags = fa->flags;
	fa->polys = poly;
	poly->numverts = lnumverts;

	for (i=0 ; i<lnumverts ; i++)
	{
		lindex = currentmodel->surfedges[fa->firstedge + i];

		if (lindex > 0)
		{
			r_pedge = &pedges[lindex];
			vec = r_pcurrentvertbase[r_pedge->v[0]].position;
		}
		else
		{
			r_pedge = &pedges[-lindex];
			vec = r_pcurrentvertbase[r_pedge->v[1]].position;
		}

		// DigitalBlur vertex light code
		// poly->verts[0...2] = world co-ordinates
		// poly->verts[3...4] = texture co-ordinates
		// poly->verts[5...6] = lightmap co-ordinates
		// poly->verts[7]     = lightmap offset in lightdata
		// poly->verts[8...10] = vertex light rgb (store it here rather than calculate it
		//    from scratch for each vert as needed cos that would be just too slow)

		// texture co-ordinates
		s = DotProduct (vec, fa->texinfo->vecs[0]) + fa->texinfo->vecs[0][3];
		t = DotProduct (vec, fa->texinfo->vecs[1]) + fa->texinfo->vecs[1][3];

		VectorCopy (vec, poly->verts[i]);
		poly->verts[i][3] = s / fa->texinfo->texture->width;
		poly->verts[i][4] = t / fa->texinfo->texture->height;

		if (fa->BadExtents) continue;

		// lightmap co-ordinates
		s -= fa->texturemins[0];
		t -= fa->texturemins[1];

		s += 8;
		t += 8;

		// lightmap offset for each vertex
		poly->verts[i][7] = (bound (01, (int) t >> 4, fa->tmax) *
			fa->smax + bound (01, (int) s >> 4, fa->smax)) * lightmap_bytes;

		s += fa->light_s * 16;
		s /= BLOCK_WIDTH * 16;

		t += fa->light_t * 16;
		t /= BLOCK_WIDTH * 16;

		poly->verts[i][5] = s;
		poly->verts[i][6] = t;
	}

	// remove co-linear points - Ed
	if (!gl_keeptjunctions.value && !(fa->flags & SURF_UNDERWATER))
	{
		for (i = 0 ; i < lnumverts ; ++i)
		{
			vec3_t v1, v2;
			float *prev, *this, *next;

			prev = poly->verts[(i + lnumverts - 1) % lnumverts];
			this = poly->verts[i];
			next = poly->verts[(i + 1) % lnumverts];

			VectorSubtract( this, prev, v1 );
			VectorNormalize( v1 );
			VectorSubtract( next, prev, v2 );
			VectorNormalize( v2 );

			// skip co-linear points
			#define COLINEAR_EPSILON 0.001

			if ((fabs( v1[0] - v2[0] ) <= COLINEAR_EPSILON) &&
				(fabs( v1[1] - v2[1] ) <= COLINEAR_EPSILON) && 
				(fabs( v1[2] - v2[2] ) <= COLINEAR_EPSILON))
			{
				int j;

				for (j = i + 1; j < lnumverts; ++j)
				{
					int k;

					for (k = 0; k < VERTEXSIZE; ++k)
						poly->verts[j - 1][k] = poly->verts[j][k];
				}
				--lnumverts;
				++nColinElim;
				// retry next vertex next time, which is now current vertex
				--i;
			}
		}
	}

	poly->numverts = lnumverts;
}

/*
========================
GL_CreateSurfaceLightmap
========================
*/
void GL_CreateSurfaceLightmap (glmsurface_t *surf)
{
	if (surf->BadExtents) return;
	if (surf->flags & SURF_DRAWSKY) return;
	if (surf->flags & SURF_DRAWTURB) return;

	surf->smax = (surf->extents[0]>>4)+1;
	surf->tmax = (surf->extents[1]>>4)+1;

	surf->lightmaptexturenum = AllocBlock (surf->smax, surf->tmax, &surf->light_s, &surf->light_t);

	surf->base = lightmaps + surf->lightmaptexturenum*lightmap_bytes*BLOCK_WIDTH*BLOCK_HEIGHT;

	surf->base += (surf->light_t * BLOCK_WIDTH + surf->light_s) * lightmap_bytes;
	R_BuildLightMap (surf, surf->base, BLOCK_WIDTH*lightmap_bytes);
}


/*
==================
GL_BuildLightmaps

Builds the lightmap texture
with all the surfaces from all brush models
==================
*/
extern qboolean hlbsp;

void GL_BuildLightmaps (void)
{
	int		i, j;
	model_t	*m;
	glmsurface_t *surf;

	memset (allocated, 0, sizeof(allocated));

	r_framecount = 1;		// no dlightcache

	if (!lightmap_textures)
	{
		lightmap_textures = texture_extension_number;
		texture_extension_number += MAX_LIGHTMAPS;
	}

	if (gl_useclight)
	{
		// fixme: gl_rgb doesn't quite work... on 3dfx anyway.  the bastards.
		gl_lightmap_format = GL_RGBA;
		lightmap_bytes = 4;
	}
	else
	{
		gl_lightmap_format = GL_LUMINANCE;
		lightmap_bytes = 1;
	}

	Con_DPrintf ("Creating Lightmaps\n");

	if (COM_CheckParm ("-neglight"))
		neglight = 0;
	else neglight = 1;

	for (j=1 ; j<MAX_MODELS ; j++)
	{
		m = cl.model_precache[j];

		if (!m) break;
		if (m->name[0] == '*') continue;

		r_pcurrentvertbase = m->vertexes;
		currentmodel = m;

		for (i=0 ; i<m->numsurfaces ; i++)
		{
			// i had this originally moved below the ifs - not that it makes much difference
			// since GL_CreateSurfaceLightmap also checks for them in it's first line.
			GL_CreateSurfaceLightmap (m->surfaces + i);

			if (m->surfaces[i].flags & SURF_DRAWTURB) continue;

			// we need polys for the sky skybox - regular sky uses warp_polys
			//if (m->surfaces[i].flags & SURF_DRAWSKY) continue;

			// this should never happen anymore, but it's left in just in case (the
			// alternative is to hang the engine)
			if (!currentmodel) Sys_Error ("Stupid Halflife BSP Error: NULL Currentmodel");

			BuildSurfaceDisplayList (m->surfaces + i);
		}
	}

 	if (gl_mtexable) SelectTextureMTex (GL_MTexture1);

	// upload all lightmaps that were filled
	Con_DPrintf ("Uploading Lightmaps\n");

	for (i=0 ; i<MAX_LIGHTMAPS ; i++)
	{
		if (!allocated[i][0])
			break;		// no more used

		lightmap_modified[i] = false;
		lightmap_rectchange[i].l = BLOCK_WIDTH;
		lightmap_rectchange[i].t = BLOCK_HEIGHT;
		lightmap_rectchange[i].w = 0;
		lightmap_rectchange[i].h = 0;

		glBindTexture (GL_TEXTURE_2D, lightmap_textures + i);
		oldtexture = lightmap_textures + i;

		glTexParameterf(GL_TEXTURE_2D, GL_TEXTURE_MIN_FILTER, GL_LINEAR);
		glTexParameterf(GL_TEXTURE_2D, GL_TEXTURE_MAG_FILTER, GL_LINEAR);

		glTexImage2D (GL_TEXTURE_2D, 0, lightmap_bytes, BLOCK_WIDTH, BLOCK_HEIGHT, 0, 
			gl_lightmap_format, GL_UNSIGNED_BYTE, 
			lightmaps + i * BLOCK_WIDTH * BLOCK_HEIGHT * lightmap_bytes);
	}

 	if (gl_mtexable) SelectTextureMTex (GL_MTexture0);

	// calculate vertex lighting data
	Con_DPrintf ("Calculating Vertex Lighting\n");

	for (j = 1; j < MAX_MODELS; j++)
	{
		m = cl.model_precache[j];

		if (!m) break;
		if (m->name[0] == '*') continue;

		currentmodel = m;

		for (i = 0; i < m->numsurfaces; i++)
		{
			if (m->surfaces[i].flags & SURF_DRAWTURB) continue;
			if (m->surfaces[i].flags & SURF_DRAWSKY) continue;

			surf = m->surfaces + i;

			if (gl_useclight)
				GL_BuildVertexLightColour (surf);
			else GL_BuildVertexLightWhite (surf);
		}
	}
}


// i had agonised for ages over how to do this, and then that bastard digitalblur goes and uses
// the code i needed for something completely different!!!!  cheers mate - i appreciate it.
void GL_BuildVertexLightColour (glmsurface_t *fa)
{
	float scale, *v;
	vec3_t vc;
	int i, size3;
	glpoly_t *p;
	byte *lm;

	if (fa->BadExtents) return;

	size3 = ((fa->extents[0] >> 4) + 1) * ((fa->extents[1] >> 4) + 1) * 4;

	for (p = fa->polys; p; p = p->next)
	{
		for (i = 0, v = p->verts[0]; i < p->numverts; i++, v += VERTEXSIZE)
		{
			vc[0] = vc[1] = vc[2] = 0;

			if (fa->styles[0] != 255)
			{
				lm = (byte *) ((long) fa->samples + (int) v[7]);
				scale = d_lightstylevalue[fa->styles[0]] * 0.0078125f;

				vc[0] += lm[0] * scale;
				vc[1] += lm[1] * scale;
				vc[2] += lm[2] * scale;

				if (fa->styles[1] != 255)
				{
					scale = d_lightstylevalue[fa->styles[1]] * 0.0078125f;

					vc[0] += lm[size3] * scale;
					vc[1] += lm[size3 + 1] * scale;
					vc[2] += lm[size3 + 2] * scale;

					if (fa->styles[2] != 255)
					{
						scale = d_lightstylevalue[fa->styles[2]] * 0.0078125f;

						vc[0] += lm[size3 * 2] * scale;
						vc[1] += lm[size3 * 2 + 1] * scale;
						vc[2] += lm[size3 * 2 + 2] * scale;

						if (fa->styles[3] != 255)
						{
							scale = d_lightstylevalue[fa->styles[3]] * 0.0078125f;

							vc[0] += lm[size3 * 3] * scale;
							vc[1] += lm[size3 * 3 + 1] * scale;
							vc[2] += lm[size3 * 3 + 2] * scale;
						}
					}
				}
			}

			// store out the vertex light colours
			p->verts[i][8] = vc[0] / 255;
			p->verts[i][9] = vc[1] / 255;
			p->verts[i][10] = vc[2] / 255;
		}
	}
}


void GL_BuildVertexLightWhite (glmsurface_t *fa)
{
	float scale, *v;
	float vc;
	int i, size3;
	glpoly_t *p;
	byte *lm;

	if (fa->BadExtents) return;

	size3 = ((fa->extents[0] >> 4) + 1) * ((fa->extents[1] >> 4) + 1);

	for (p = fa->polys; p; p = p->next)
	{
		for (i = 0, v = p->verts[0]; i < p->numverts; i++, v += VERTEXSIZE)
		{
			vc = 0;

			if (fa->styles[0] != 255)
			{
				lm = (byte *) ((long) fa->samples + (int) v[7]);
				scale = d_lightstylevalue[fa->styles[0]] * 0.0078125f;
				vc += lm[0] * scale;

				if (fa->styles[1] != 255)
				{
					scale = d_lightstylevalue[fa->styles[1]] * 0.0078125f;
					vc += lm[size3] * scale;

					if (fa->styles[2] != 255)
					{
						scale = d_lightstylevalue[fa->styles[2]] * 0.0078125f;
						vc += lm[size3 * 2] * scale;

						if (fa->styles[3] != 255)
						{
							scale = d_lightstylevalue[fa->styles[3]] * 0.0078125f;
							vc += lm[size3 * 3] * scale;
						}
					}
				}
			}

			// store out the vertex light colours
			p->verts[i][8] = vc / 255;
			p->verts[i][9] = vc / 255;
			p->verts[i][10] = vc / 255;
		}
	}
}
